import pytest, warnings, numpy as np
from sequentia.classifiers import _Topology, _LeftRightTopology, _ErgodicTopology, _LinearTopology
from ....support import assert_equal, assert_all_equal, assert_distribution

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# ========= #
# _Topology #
# ========= #

# --------------------------- #
# _Topology.uniform_initial() #
# --------------------------- #

def test_uniform_initial_min():
    """Generate a uniform initial state distribution with the minimum number of states"""
    topology = _Topology(n_states=1, random_state=rng)
    initial = topology.uniform_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        1.
    ]))

def test_uniform_initial_small():
    """Generate a uniform initial state distribution with a few states"""
    topology = _Topology(n_states=2, random_state=rng)
    initial = topology.uniform_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.5, 0.5
    ]))

def test_uniform_initial_many():
    """Generate a uniform initial state distribution with many states"""
    topology = _Topology(n_states=5, random_state=rng)
    initial = topology.uniform_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

# -------------------------- #
# _Topology.random_initial() #
# -------------------------- #

def test_random_initial_min():
    """Generate a random initial state distribution with minimal states"""
    topology = _Topology(n_states=1, random_state=rng)
    initial = topology.random_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        1.
    ]))

def test_random_initial_small():
    """Generate a random initial state distribution with few states"""
    topology = _Topology(n_states=2, random_state=rng)
    initial = topology.random_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.57633871, 0.42366129
    ]))

def test_random_initial_many():
    """Generate a random initial state distribution with many states"""
    topology = _Topology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.15210286, 0.10647349, 0.20059295, 0.11120171, 0.42962898
    ]))

# ================== #
# _LeftRightTopology #
# ================== #

# ---------------------------------------- #
# _LeftRightTopology.uniform_transitions() #
# ---------------------------------------- #

def test_left_right_uniform_transitions_min():
    """Generate a uniform left-right transition matrix with minimal states"""
    topology = _LeftRightTopology(n_states=1, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_left_right_uniform_transitions_small():
    """Generate a uniform left-right transition matrix with few states"""
    topology = _LeftRightTopology(n_states=2, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5],
        [0. , 1. ]
    ]))

def test_left_right_uniform_transitions_many():
    """Generate a uniform left-right transition matrix with many states"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.2, 0.2 , 0.2       , 0.2       , 0.2       ],
        [0. , 0.25, 0.25      , 0.25      , 0.25      ],
        [0. , 0.  , 0.33333333, 0.33333333, 0.33333333],
        [0. , 0.  , 0.        , 0.5       , 0.5       ] ,
        [0. , 0.  , 0.        , 0.        , 1.        ]
    ]))

# --------------------------------------- #
# _LeftRightTopology.random_transitions() #
# --------------------------------------- #

def test_left_right_random_transitions_min():
    """Generate a random left-right transition matrix with minimal states"""
    topology = _LeftRightTopology(n_states=1, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_left_right_random_transitions_small():
    """Generate a random left-right transition matrix with few states"""
    topology = _LeftRightTopology(n_states=2, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.23561633, 0.76438367],
        [0.        , 1.        ]
    ]))

def test_left_right_random_transitions_many():
    """Generate a random left-right transition matrix with many states"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.23169814, 0.71716356, 0.02033845, 0.02516204, 0.00563782],
        [0.        , 0.19474072, 0.16405008, 0.22228532, 0.41892388],
        [0.        , 0.        , 0.42912755, 0.16545797, 0.40541448],
        [0.        , 0.        , 0.        , 0.109713  , 0.890287  ],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

# ----------------------------------------- #
# _LeftRightTopology.validate_transitions() #
# ----------------------------------------- #

def test_left_right_validate_transitions_invalid():
    """Validate an invalid left-right transition matrix"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = _ErgodicTopology(n_states=5, random_state=rng).random_transitions()
    with pytest.raises(ValueError) as e:
        topology.validate_transitions(transitions)
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_left_right_validate_transitions_valid():
    """Validate a valid left-right transition matrix"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    topology.validate_transitions(transitions)

# -------------------------------------- #
# _ErgodicTopology.uniform_transitions() #
# -------------------------------------- #

def test_ergodic_uniform_transitions_min():
    """Generate a uniform ergodic transition matrix with minimal states"""
    topology = _ErgodicTopology(n_states=1, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_ergodic_uniform_transitions_small():
    """Generate a uniform ergodic transition matrix with few states"""
    topology = _ErgodicTopology(n_states=2, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5],
        [0.5, 0.5]
    ]))

def test_ergodic_uniform_transitions_many():
    """Generate a uniform ergodic transition matrix with many states"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))

# ------------------------------------- #
# _ErgodicTopology.random_transitions() #
# ------------------------------------- #

def test_ergodic_random_transitions_min():
    """Generate a random ergodic transition matrix with minimal states"""
    topology = _ErgodicTopology(n_states=1, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_ergodic_random_transitions_small():
    """Generate a random ergodic transition matrix with few states"""
    topology = _ErgodicTopology(n_states=2, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.9474011 , 0.0525989 ],
        [0.85567599, 0.14432401]
    ]))

def test_ergodic_random_transitions_many():
    """Generate a random ergodic transition matrix with many states"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.58715548, 0.14491542, 0.20980762, 0.00623944, 0.05188205],
        [0.0840705 , 0.23055049, 0.08297536, 0.25124688, 0.35115677],
        [0.02117615, 0.37664662, 0.26705912, 0.09851123, 0.23660688],
        [0.01938041, 0.16853843, 0.52046123, 0.07535256, 0.21626737],
        [0.04996846, 0.44545843, 0.12079423, 0.07154241, 0.31223646]
    ]))

# --------------------------------------- #
# _ErgodicTopology.validate_transitions() #
# --------------------------------------- #

def test_ergodic_validate_transitions_invalid():
    """Validate an invalid ergodic transition matrix"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = _LeftRightTopology(n_states=5, random_state=rng).random_transitions()
    with pytest.warns(UserWarning):
        topology.validate_transitions(transitions)

def test_ergodic_validate_transitions_valid():
    """Validate a valid ergodic transition matrix"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    topology.validate_transitions(transitions)

# =============== #
# _LinearTopology #
# =============== #

# ------------------------------------- #
# _LinearTopology.uniform_transitions() #
# ------------------------------------- #

def test_linear_uniform_transitions_min():
    """Generate a uniform linear transition matrix with minimal states"""
    topology = _LinearTopology(n_states=1, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_linear_uniform_transitions_small():
    """Generate a uniform linear transition matrix with few states"""
    topology = _LinearTopology(n_states=2, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5],
        [0. , 1. ]
    ]))

def test_linear_uniform_transitions_many():
    """Generate a uniform linear transition matrix with many states"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5, 0. , 0. , 0. ],
        [0. , 0.5, 0.5, 0. , 0. ],
        [0. , 0. , 0.5, 0.5, 0. ],
        [0. , 0. , 0. , 0.5, 0.5],
        [0. , 0. , 0. , 0. , 1. ]
    ]))

# ------------------------------------ #
# _LinearTopology.random_transitions() #
# ------------------------------------ #

def test_linear_random_transitions_min():
    """Generate a random linear transition matrix with minimal states"""
    topology = _LinearTopology(n_states=1, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_linear_random_transitions_small():
    """Generate a random linear transition matrix with few states"""
    topology = _LinearTopology(n_states=2, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.65157396, 0.34842604],
        [0.        , 1.        ]
    ]))

def test_linear_random_transitions_many():
    """Generate a random linear transition matrix with many states"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.44455421, 0.55544579, 0.        , 0.        , 0.        ],
        [0.        , 0.57553614, 0.42446386, 0.        , 0.        ],
        [0.        , 0.        , 0.92014965, 0.07985035, 0.        ],
        [0.        , 0.        , 0.        , 0.66790982, 0.33209018],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

# -------------------------------------- #
# _LinearTopology.validate_transitions() #
# -------------------------------------- #

def test_linear_validate_transitions_invalid():
    """Validate an invalid linear transition matrix"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = _ErgodicTopology(n_states=5, random_state=rng).random_transitions()
    with pytest.raises(ValueError) as e:
        topology.validate_transitions(transitions)
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_linear_validate_transitions_valid():
    """Validate a valid linear transition matrix"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    topology.validate_transitions(transitions)


import numpy as np
from ..internals import _Validator

__all__ = ['Transform', 'Custom', 'TrimConstants', 'MinMaxScale', 'Center', 'Standardize', 'Downsample', 'Filter']

class Transform:
    """Base class representing a single transformation."""

    def __init__(self):
        self._val = _Validator()
        self._name = self.__class__.__name__

    def __call__(self, X, validate=True):
        """Applies the transformation to the observation sequence(s).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        validate: bool
            Whether or not to validate the input sequences.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float) or list of :class:`numpy:numpy.ndarray` (float)
            The transformed input observation sequence(s).
        """
        if self._val.is_boolean(validate, 'validate'):
            X = self._val.is_observation_sequences(X, allow_single=True)

        if self.is_fitted():
            return self._apply(X)

        try:
            self.fit(X, validate=validate)
            return self._apply(X)
        except:
            raise
        finally:
            self.unfit()

    def transform(self, x):
        """Applies the transformation to a single observation sequence.

        Parameters
        ----------
        X: numpy.ndarray (float)
            An individual observation sequence.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float)
            The transformed input observation sequence.
        """
        raise NotImplementedError

    def fit(self, X, validate=True):
        """Fit the transformation on the provided observation sequence(s) (without transforming them).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        validate: bool
            Whether or not to validate the input sequences.
        """
        self._val.is_boolean(validate, 'validate')

    def fit_transform(self, X, validate=True):
        """Fit the transformation with the provided observation sequence(s) and transform them.

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        validate: bool
            Whether or not to validate the input sequences.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float) or list of :class:`numpy:numpy.ndarray` (float)
            The transformed input observation sequence(s).
        """
        self.fit(X, validate=validate)
        return self.__call__(X, validate=validate)

    def is_fitted(self):
        """Check whether or not the transformation is fitted on some observation sequence(s).

        Returns
        -------
        fitted: bool
            Whether or not the transformation is fitted.
        """
        return False

    def unfit(self):
        """Unfit the transformation by resetting the parameters to their default settings."""
        pass

    def _apply(self, X):
        """Applies the transformation to the observation sequence(s) (internal).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float) or list of :class:`numpy:numpy.ndarray` (float)
            The transformed input observation sequence(s).
        """
        return self.transform(X) if isinstance(X, np.ndarray) else [self.transform(x) for x in X]

    def __str__(self):
        """Description of the transformation.

        Returns
        -------
        description: str
            The description of the transformation.
        """
        raise NotImplementedError

class Custom(Transform):
    """Apply a custom transformation to the input observation sequence(s).

    Parameters
    ----------
    func: callable
        A lambda or function that specifies the transformation that should be applied to a **single** observation sequence.

    name: str
        Name of the transformation.

    desc: str
        Description of the transformation.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Apply a custom transformation
    >>> X = Custom(lambda x: x**2, name='Square', desc='Square observations element-wise')(X)
    """

    def __init__(self, func, name=None, desc=None):
        super().__init__()
        self.transform = self._val.is_func(func, 'transformation')
        self._name = 'Custom' + ('' if name is None else ' ({})'.format(self._val.is_string(name, 'name')))
        self._desc = 'Apply a custom transformation' if desc is None else self._val.is_string(desc, 'description')

    def __str__(self):
        return self._desc

class TrimConstants(Transform):
    """Trim constant observations from the input observation sequence(s).

    Parameters
    ----------
    const: float
        The constant value.

    Examples
    --------
    >>> # Create some sample data
    >>> z = np.zeros((4, 3))
    >>> x = lambda i: np.vstack((z, np.random.random((10 * i, 3)), z))
    >>> X = [x(i) for i in range(1, 4)]
    >>> # Trim the data
    >>> X = TrimConstants()(X)
    """

    def __init__(self, constant=0):
        super().__init__()
        try:
            self.constant = float(constant)
        except ValueError:
            raise TypeError('Expected constant to be a float or convertible to a float')

    def transform(self, x):
        return x[~np.all(x == self.constant, axis=1)]

    def __str__(self):
        return 'Remove constant observations (={:.3})'.format(self.constant)

class MinMaxScale(Transform):
    """Scales the observation sequence features to each be within a provided range.

    Parameters
    ----------
    scale: tuple(int/float, int/float)
        The range of the transformed observation sequence features.

    independent: bool
        Whether to independently compute the minimum and maximum to scale each observation sequence.
    """

    def __init__(self, scale=(0., 1.), independent=True):
        super().__init__()
        if not isinstance(scale, tuple):
            raise TypeError('Expected scaling range to be a tuple')
        if not all(isinstance(val, (int, float)) for val in scale):
            raise TypeError('Expected upper and lower bounds of scaling range to be floats')
        if not scale[0] < scale[1]:
            raise ValueError('Expected lower bound of scaling range to be less than the upper bound')
        self.scale = scale
        self.independent = self._val.is_boolean(independent, 'independent')
        self._type = (_MinMaxScaleIndependent if independent else _MinMaxScaleAll)(scale)

    def fit(self, X, validate=True):
        super().fit(X, validate=validate)
        self._type.fit(X, validate=validate)

    def transform(self, x):
        return self._type.transform(x)

    def is_fitted(self):
        return self._type.is_fitted()

    def unfit(self):
        self._type.unfit()

    def __str__(self):
        return str(self._type)

class _MinMaxScaleAll(Transform):
    def __init__(self, scale):
        super().__init__()
        self.scale = scale
        self.unfit()

    def fit(self, X, validate):
        if validate:
            X = self._val.is_observation_sequences(X, allow_single=True)
        if isinstance(X, list):
            X = np.vstack(X)
        self.min, self.max = X.min(axis=0), X.max(axis=0)

    def transform(self, x):
        min_, max_ = self.scale
        scale = (max_ - min_) / (self.max - self.min)
        return scale * x + min_ - self.min * scale

    def is_fitted(self):
        return (self.min is not None) and (self.max is not None)

    def unfit(self):
        self.min, self.max = None, None

    def __str__(self):
        return 'Min-max scaling into range {} (all)'.format(self.scale)

class _MinMaxScaleIndependent(Transform):
    def __init__(self, scale):
        super().__init__()
        self.scale = scale

    def transform(self, x):
        min_, max_ = self.scale
        scale = (max_ - min_) / (x.max(axis=0) - x.min(axis=0))
        return scale * x + min_ - x.min(axis=0) * scale

    def __str__(self):
        return 'Min-max scaling into range {} (independent)'.format(self.scale)

class Center(Transform):
    """Centers the observation sequence features around their means. Results in zero-mean features.

    Parameters
    ----------
    independent: bool
        Whether to independently compute the mean to scale each observation sequence.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Center the data
    >>> X = Center()(X)
    """

    def __init__(self, independent=True):
        super().__init__()
        self.independent = self._val.is_boolean(independent, 'independent')
        self._type = (_CenterIndependent if independent else _CenterAll)()

    def fit(self, X, validate=True):
        super().fit(X, validate=validate)
        self._type.fit(X, validate=validate)

    def transform(self, x):
        return self._type.transform(x)

    def is_fitted(self):
        return self._type.is_fitted()

    def unfit(self):
        self._type.unfit()

    def __str__(self):
        return str(self._type)

class _CenterAll(Transform):
    def __init__(self):
        super().__init__()
        self.unfit()

    def fit(self, X, validate):
        if validate:
            X = self._val.is_observation_sequences(X, allow_single=True)
        if isinstance(X, list):
            X = np.vstack(X)
        self.mean = X.mean(axis=0)

    def transform(self, x):
        return x - self.mean

    def is_fitted(self):
        return self.mean is not None

    def unfit(self):
        self.mean = None

    def __str__(self):
        return 'Centering around mean (zero mean) (all)'

class _CenterIndependent(Transform):
    def transform(self, x):
        return x - x.mean(axis=0)

    def __str__(self):
        return 'Centering around mean (zero mean) (independent)'

class Standardize(Transform):
    """Centers the observation sequence features around their means, then scales them by their deviations.
    Results in zero-mean, unit-variance features.

    Parameters
    ----------
    independent: bool
        Whether to independently compute the mean and standard deviation to scale each observation sequence.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Standardize the data
    >>> X = Standardize()(X)
    """

    def __init__(self, independent=True):
        super().__init__()
        self.independent = self._val.is_boolean(independent, 'independent')
        self._type = (_StandardizeIndependent if independent else _StandardizeAll)()

    def fit(self, X, validate=True):
        super().fit(X, validate=validate)
        self._type.fit(X, validate=validate)

    def transform(self, x):
        return self._type.transform(x)

    def is_fitted(self):
        return self._type.is_fitted()

    def unfit(self):
        self._type.unfit()

    def __str__(self):
        return str(self._type)

class _StandardizeAll(Transform):
    def __init__(self):
        super().__init__()
        self.unfit()

    def fit(self, X, validate):
        if validate:
            X = self._val.is_observation_sequences(X, allow_single=True)
        if isinstance(X, list):
            X = np.vstack(X)
        self.mean, self.std = X.mean(axis=0), X.std(axis=0)

    def transform(self, x):
        return (x - self.mean) / self.std

    def is_fitted(self):
        return (self.mean is not None) and (self.std is not None)

    def unfit(self):
        self.mean, self.std = None, None

    def __str__(self):
        return 'Standard scaling (zero mean, unit variance) (all)'

class _StandardizeIndependent(Transform):
    def transform(self, x):
        return (x - x.mean(axis=0)) / x.std(axis=0)

    def __str__(self):
        return 'Standard scaling (zero mean, unit variance) (independent)'

class Downsample(Transform):
    """Downsamples an observation sequence (or multiple sequences) by either:

    - Decimating the next :math:`n-1` observations
    - Averaging the current observation with the next :math:`n-1` observations

    Parameters
    ----------
    factor: int > 0
        Downsample factor.

    method: {'decimate', 'mean'}
        The downsampling method.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Downsample the data with downsample factor 5 and decimation
    >>> X = Downsample(factor=5, method='decimate')(X)
    """

    def __init__(self, factor, method='decimate'):
        super().__init__()
        self.factor = self._val.is_restricted_integer(factor, lambda x: x > 0, desc='downsample factor', expected='positive')
        self.method = self._val.is_one_of(method, ['decimate', 'mean'], desc='downsampling method')
        self._type = (_DownsampleDecimate if method == 'decimate' else _DownsampleMean)(factor)

    def transform(self, x):
        return self._type.transform(x)

    def _apply(self, X):
        if isinstance(X, np.ndarray):
            self._val.is_restricted_integer(self.factor, lambda x: x <= len(X),
                desc='downsample factor', expected='no greater than the number of frames')
        else:
            self._val.is_restricted_integer(self.factor, lambda x: x <= min(len(x) for x in X),
                desc='downsample factor', expected='no greater than the number of frames in the shortest sequence')
        return super()._apply(X)

    def __str__(self):
        return str(self._type)

class _DownsampleDecimate(Transform):
    def __init__(self, factor):
        self.factor = factor

    def transform(self, x):
        return np.delete(x, [i for i in range(len(x)) if i % self.factor != 0], 0)

    def __str__(self):
        return 'Decimation downsampling with factor {}'.format(self.factor)

class _DownsampleMean(Transform):
    def __init__(self, factor):
        self.factor = factor

    def transform(self, x):
        N, D = x.shape
        r = len(x) % self.factor
        xn, xr = (x, None) if r == 0 else (x[:-r], x[-r:])
        dxn = xn.T.reshape(-1, self.factor).mean(axis=1).reshape(D, -1).T
        return dxn if xr is None else np.vstack((dxn, xr.mean(axis=0)))

    def __str__(self):
        return 'Mean downsampling with factor {}'.format(self.factor)

class Filter(Transform):
    """Applies a median or mean filter to the input observation sequence(s).

    Parameters
    ----------
    window_size: int > 0
        The size of the filtering window.

    method: {'median', 'mean'}
        The filtering method.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Filter the data with window size 5 and median filtering
    >>> X = Filter(window_size=5, method='median')(X)
    """

    def __init__(self, window_size, method='median'):
        super().__init__()
        self.window_size = self._val.is_restricted_integer(window_size, lambda x: x > 0, desc='window size', expected='positive')
        self.method = self._val.is_one_of(method, ['median', 'mean'], desc='filtering method')
        self._func = np.median if self.method == 'median' else np.mean

    def transform(self, x):
        filtered = []
        right = self.window_size // 2
        left = (self.window_size - 1) - right
        for i in range(len(x)):
            l, m, r = x[((i - left) * (left < i)):i], x[i], x[(i + 1):(i + 1 + right)]
            filtered.append(self._func(np.vstack((l, m, r)), axis=0))
        return np.array(filtered)

    def _apply(self, X):
        if isinstance(X, np.ndarray):
            self._val.is_restricted_integer(self.window_size, lambda x: x <= len(X),
                desc='window size', expected='no greater than the number of frames')
        else:
            self._val.is_restricted_integer(self.window_size, lambda x: x <= min(len(x) for x in X),
                desc='window size', expected='no greater than the number of frames in the shortest sequence')
        return super()._apply(X)

    def __str__(self):
        return '{} filtering with window-size {}'.format(self.method.capitalize(), self.window_size)


import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

class DeepGRU(nn.Module):
    """A modular PyTorch implementation of the DeepGRU (Deep Gesture Recognition Utility) recurrent neural network architecture
    designed by Maghoumi & LaViola Jr. [#maghoumi]_, originally for gesture recognition, but applicable to general sequence classification tasks.

    Parameters
    ----------
    n_features: int
        The number of features that each observation within a sequence has.

    n_classes: int
        The number of different sequence classes.

    dims: dict
        A dictionary consisting of dimension configuration for the GRUs and fully-connected layers.

        .. note::
            Values for the keys ``'gru1'``, ``'gru2'``, ``'gru3'`` and ``'fc'`` must be set.

    device: str, optional
        The device to send the model parameters to for computation.

        If no device is specified, a check is made for any available CUDA device, otherwise the CPU is used.

    Notes
    -----
    .. [#maghoumi] **Mehran Maghoumi & Joseph J. LaViola Jr.** `"DeepGRU: Deep Gesture Recognition Utility" <https://arxiv.org/abs/1810.12514>`_
      *Advances in Visual Computing, 14th International Symposium on Visual Computing*, ISVC 2019,
      Lake Tahoe, NV, USA, October 7–9, 2019, Proceedings, Part I (pp.16-31)
    """
    def __init__(self, n_features, n_classes, dims={'gru1': 512, 'gru2': 256, 'gru3': 128, 'fc': 256}, device=None):
        super().__init__()

        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'

        # Specify sub-modules
        self.model = nn.ModuleDict({
            'enc': _EncoderNetwork(dims={'in': n_features, **{k:v for k, v in dims.items() if k.startswith('gru')}}, device=device),
            'attn': _AttentionModule(dims={'in': dims['gru3']}, device=device),
            'clf': _Classifier(dims={'in': dims['gru3']*2, 'fc': dims['fc'], 'out': n_classes}, device=device),
        })

        # Send model to device
        self.to(device)

    def forward(self, x, x_lengths):
        """Passes the batched input sequences through the encoder network, attention module and classifier to generate log-softmax scores.

        .. note::
            Since log-softmax scores are returned, it is advised to use the negative log-likelihood loss :class:`torch:torch.nn.NLLLoss`.

        Parameters
        ----------
        x: torch.PackedSequence
            A packed representation of a batch of input observation sequences.

        x_lengths: torch.Tensor (long/int)
            A tensor of the sequence lengths of the batch in descending order.

        Returns
        -------
        log_softmax: :class:`torch:torch.Tensor` (float)
            :math:`B\\times C` tensor of :math:`C` log-softmax scores (class predictions) for each observation sequence in the batch.
        """
        h, h_last = self.model['enc'](x, x_lengths)
        o_attn = self.model['attn'](h, h_last)
        return self.model['clf'](o_attn)

class _EncoderNetwork(nn.Module):
    def __init__(self, dims, device):
        super().__init__()
        self.dims = dims
        self.device = device

        # Specify sub-modules
        self.model = nn.ModuleDict({
           'gru1': nn.GRU(self.dims['in'], self.dims['gru1'], num_layers=2, batch_first=True).to(device),
           'gru2': nn.GRU(self.dims['gru1'], self.dims['gru2'], num_layers=2, batch_first=True).to(device),
           'gru3': nn.GRU(self.dims['gru2'], self.dims['gru3'], num_layers=1, batch_first=True).to(device)
        })

        # Send model to device
        self.to(device)

    def forward(self, x, x_lengths):
        x = x.to(self.device)

        # Pack the padded Tensor into a PackedSequence
        x_packed = pack_padded_sequence(x, x_lengths.cpu(), batch_first=True)

        # Pass the PackedSequence through the GRUs
        h_packed, _ = self.model['gru1'](x_packed)
        h_packed, _ = self.model['gru2'](h_packed)
        h_packed, h_last = self.model['gru3'](h_packed)

        # Unpack the hidden state PackedSequence into a padded Tensor
        h_padded = pad_packed_sequence(h_packed, batch_first=True, padding_value=0.0, total_length=max(x_lengths))
        return h_padded[0], h_last
        # Shape: B x T_max x D_out, 1 x B x D_out

class _AttentionModule(nn.Module):
    def __init__(self, dims, device):
        super().__init__()
        self.device = device
        self.dims = dims

        # Specify sub-modules
        self.model = nn.ModuleDict({
            # Attentional context vector weights
            'attn_ctx': nn.Linear(self.dims['in'], self.dims['in'], bias=False).to(device),
            # Auxilliary context
            'aux_ctx': nn.GRU(input_size=self.dims['in'], hidden_size=self.dims['in']).to(device)
        })

        # Send model to device
        self.to(device)

    def forward(self, h, h_last):
        h_last.transpose_(1, 0)
        # Shape: B x 1 x D_out

        # Calculate attentional context
        h.transpose_(1, 2)
        c = F.softmax(self.model['attn_ctx'](h_last) @ h, dim=0)
        c = (c @ h.transpose(2, 1)).transpose(1, 0)
        # Shape: 1 x B x D_out

        # Calculate auxilliary context
        c_aux, _ = self.model['aux_ctx'](c, h_last.transpose(1, 0))
        # Shape: 1 x B x D_out

        # Combine attentional and auxilliary context
        return torch.cat((c.squeeze(0), c_aux.squeeze(0)), dim=1)
        # Shape: B x D_out*2

class _Classifier(nn.Module):
    def __init__(self, dims, device):
        super().__init__()
        self.device = device
        self.dims = dims

        # Specify sub-modules
        self.model = nn.ModuleDict({
            'fc1': nn.Sequential(
                nn.BatchNorm1d(self.dims['in']),
                nn.Dropout(),
                nn.Linear(self.dims['in'], self.dims['fc'])
            ).to(device),
            'fc2': nn.Sequential(
                nn.BatchNorm1d(self.dims['fc']),
                nn.Dropout(),
                nn.Linear(self.dims['fc'], self.dims['out'])
            ).to(device)
        })

        # Send model to device
        self.to(device)

    def forward(self, o_attn):
        f1 = self.model['fc1'](o_attn)
        f2 = self.model['fc2'](F.relu(f1))
        return F.log_softmax(f2, dim=1)


import importlib, platform
from pkg_resources import packaging
if packaging.version.parse(platform.python_version()) < packaging.version.parse('3.8'):
    import importlib_metadata as metadata
else:
    from importlib import metadata

MIN_TORCH_VERSION = '1.8'

def check_package(pkg, min_version, url):
    """Checks whether a specified package has been installed,
    and whether the installed version meets a specified minimum.

    Parameters
    ----------
    pkg: str
        Name of the package.

    min_version: str
        Minimum version for the package, e.g. `1.8`.

    url: str
        Package installation page URL (for help).
    """
    try:
        importlib.import_module(pkg)
    except ImportError:
        msg = ("Could not find a valid installation of '{pkg}' (>={min_version}), which Sequentia depends on.\n"
        "Visit {url} for more instructions on installing this package.").format(pkg=pkg, url=url, min_version=min_version)
        raise ModuleNotFoundError(msg)

    installed_version = metadata.version(pkg)
    if packaging.version.parse(installed_version) < packaging.version.parse(min_version):
        msg = ("Could not find a compatible installation of '{pkg}' (>={min_version}), which Sequentia depends on - got version {installed_version}.\n"
        "Visit {url} for more instructions on installing this package.").format(pkg=pkg, url=url, min_version=min_version, installed_version=installed_version)
        raise ImportWarning(msg)

# Check that (at least) the minimum dependency version is installed
check_package('torch', MIN_TORCH_VERSION, url='https://pytorch.org/')

# Import from the package
from .deepgru import DeepGRU, _EncoderNetwork, _AttentionModule, _Classifier
from .utils import collate_fn


import torch

def collate_fn(batch):
    """Collects together univariate or multivariate sequences into a single batch, arranged in descending order of length.

    Also returns the corresponding lengths and labels as :class:`torch:torch.LongTensor` objects.

    Parameters
    ----------
    batch: list of tuple(torch.FloatTensor, int)
        Collection of :math:`B` sequence-label pairs, where the :math:`n^\\text{th}` sequence is of shape :math:`(T_n \\times D)` or :math:`(T_n,)` and the label is an integer.

    Returns
    -------
    padded_sequences: :class:`torch:torch.Tensor` (float)
        A tensor of size :math:`B \\times T_\\text{max} \\times D` containing all of the sequences in descending length order, padded to the length of the longest sequence in the batch.

    lengths: :class:`torch:torch.Tensor` (long/int)
        A tensor of the :math:`B` sequence lengths in descending order.

    labels: :class:`torch:torch.Tensor` (long/int)
        A tensor of the :math:`B` sequence labels in descending length order.
    """
    batch_size = len(batch)

    # Sort the (sequence, label) pairs in descending order of duration
    batch.sort(key=(lambda x: len(x[0])), reverse=True)
    # Shape: list(tuple(tensor(TxD), int)) or list(tuple(tensor(T), int))

    # Create list of sequences, and tensors for lengths and labels
    sequences, lengths, labels = [], torch.zeros(batch_size, dtype=torch.long), torch.zeros(batch_size, dtype=torch.long)
    for i, (sequence, label) in enumerate(batch):
        lengths[i], labels[i] = len(sequence), label
        sequences.append(sequence)

    # Combine sequences into a padded matrix
    padded_sequences = torch.nn.utils.rnn.pad_sequence(sequences, batch_first=True)
    # Shape: (B x T_max x D) or (B x T_max)

    # If a vector input was given for the sequences, expand (B x T_max) to (B x T_max x 1)
    if padded_sequences.ndim == 2:
        padded_sequences.unsqueeze_(-1)

    return padded_sequences, lengths, labels
    # Shapes: (B x T_max x D), (B,), (B,)


import re, numpy as np, hmmlearn.hmm
from .topologies.ergodic import _ErgodicTopology
from .topologies.left_right import _LeftRightTopology
from .topologies.linear import _LinearTopology
from ...internals import _Validator

class GMMHMM:
    """A hidden Markov model (with Gaussian Mixture Model emissions)
    representing a single isolated sequence class.

    Parameters
    ----------
    label: str or numeric
        A label for the model, corresponding to the class being represented.

    n_states: int > 0
        The number of states for the model.

    n_components: int > 0
        The number of mixture components used in the emission distribution for each state.

    covariance_type: {'spherical', 'diag', 'full', 'tied'}
        The covariance matrix type for emission distributions.

    topology: {'ergodic', 'left-right', 'linear'}
        The topology for the model.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Attributes
    ----------
    label (property): str or numeric
        The label for the model.

    model (property): hmmlearn.hmm.GMMHMM
        The underlying GMMHMM model from `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/api.html#gmmhmm>`_.

    n_states (property): int
        The number of states for the model.

    n_components (property): int
        The number of mixture components used in the emission distribution for each state.

    covariance_type (property): str
        The covariance matrix type for emission distributions.

    frozen (property): set (str)
        The frozen parameters of the HMM or its GMM emission distributions (see :func:`freeze`).

    n_seqs_ (property): int
        The number of observation sequences used to train the model.

    initial_ (property/setter): numpy.ndarray (float)
        The initial state distribution of the model.

    transitions_ (property/setter): numpy.ndarray (float)
        The transition matrix of the model.

    weights_ (property): numpy.ndarray (float)
        The mixture weights of the GMM emission distributions.

    means_ (property): numpy.ndarray (float)
        The mean vectors of the GMM emission distributions.

    covars_ (property): numpy.ndarray (float)
        The covariance matrices of the GMM emission distributions.

    monitor_ (property): hmmlearn.base.ConvergenceMonitor
        The convergence monitor for the Baum–Welch algorithm.
    """

    def __init__(self, label, n_states, n_components=1, covariance_type='full', topology='left-right', random_state=None):
        self._val = _Validator()
        self._label = self._val.is_string_or_numeric(label, 'model label')
        self._label = label
        self._n_states = self._val.is_restricted_integer(
            n_states, lambda x: x > 0, desc='number of states', expected='greater than zero')
        self._n_components = self._val.is_restricted_integer(
            n_components, lambda x: x > 0, desc='number of mixture components', expected='greater than zero')
        self._covariance_type = self._val.is_one_of(covariance_type, ['spherical', 'diag', 'full', 'tied'], desc='covariance matrix type')
        self._val.is_one_of(topology, ['ergodic', 'left-right', 'linear'], desc='topology')
        self._random_state = self._val.is_random_state(random_state)
        self._topology = {
            'ergodic': _ErgodicTopology,
            'left-right': _LeftRightTopology,
            'linear': _LinearTopology
        }[topology](self._n_states, self._random_state)
        self._frozen = set()

    def set_uniform_initial(self):
        """Sets a uniform initial state distribution :math:`\\boldsymbol{\\pi}=(\\pi_1,\\pi_2,\\ldots,\\pi_M)` where :math:`\\pi_i=1/M\\quad\\forall i`."""
        self._initial_ = self._topology.uniform_initial()

    def set_random_initial(self):
        """Sets a random initial state distribution by sampling :math:`\\boldsymbol{\\pi}\\sim\\mathrm{Dir}(\\mathbf{1}_M)` where

        - :math:`\\boldsymbol{\\pi}=(\\pi_1,\\pi_2,\\ldots,\\pi_M)` are the initial state probabilities for each state,
        - :math:`\\mathbf{1}_M` is a vector of :math:`M` ones which are used as the concentration parameters for the Dirichlet distribution.
        """
        self._initial_ = self._topology.random_initial()

    def set_uniform_transitions(self):
        """Sets a uniform transition matrix according to the topology, so that given the HMM is in state :math:`i`,
        all permissible transitions (i.e. such that :math:`p_{ij}\\neq0`) :math:`\\forall j` are equally probable."""
        self._transitions_ = self._topology.uniform_transitions()

    def set_random_transitions(self):
        """Sets a random transition matrix according to the topology, so that given the HMM is in state :math:`i`,
        all out-going transition probabilities :math:`\\mathbf{p}_i=(p_{i1},p_{i2},\\ldots,p_{iM})` from state :math:`i`
        are generated by sampling :math:`\\mathbf{p}_i\\sim\\mathrm{Dir}(\\mathbf{1})` with a vector of ones of appropriate
        size used as concentration parameters, so that only transitions permitted by the topology are non-zero."""
        self._transitions_ = self._topology.random_transitions()

    def fit(self, X):
        """Fits the HMM to observation sequences assumed to be labeled as the class that the model represents.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            Collection of multivariate observation sequences, each of shape :math:`(T \\times D)` where
            :math:`T` may vary per observation sequence.
        """
        (self.initial_, self.transitions_)
        X = self._val.is_observation_sequences(X)

        # Store the number of sequences and features used to fit the model
        self._n_seqs_, self._n_features_ = len(X), X[0].shape[1]

        # Initialize the GMMHMM with the specified initial state distribution and transition matrix
        self._model = hmmlearn.hmm.GMMHMM(
            n_components=self._n_states,
            n_mix=self._n_components,
            covariance_type=self._covariance_type,
            random_state=self._random_state,
            init_params='mcw', # only initialize means, covariances and mixture weights
            params=(set('stmcw') - self._frozen)
        )
        self._model.startprob_, self._model.transmat_ = self._initial_, self._transitions_

        # Perform the Baum-Welch algorithm to fit the model to the observations
        self._model.fit(np.vstack(X), [len(x) for x in X])

        # Update the initial state distribution and transitions to reflect the updated parameters
        self._initial_, self._transitions_ = self._model.startprob_, self._model.transmat_

    def forward(self, x):
        """Runs the forward algorithm to calculate the (log) likelihood of the model generating an observation sequence.

        Parameters
        ----------
        x: numpy.ndarray (float)
            An individual sequence of observations of size :math:`(T \\times D)` where
            :math:`T` is the number of time frames (or observations) and
            :math:`D` is the number of features.

        Returns
        -------
        log-likelihood: float
            The log-likelihood of the model generating the observation sequence.
        """
        self.model
        x = self._val.is_observation_sequences(x, allow_single=True)
        if not x.shape[1] == self._n_features_:
            raise ValueError('Number of observation features must match the dimensionality of the original data used to fit the model')
        return self._model.score(x, lengths=None)

    def freeze(self, params=None):
        """Freezes the specified parameters of the HMM or its GMM emission distributions,
        preventing them from being updated during the Baum–Welch algorithm.

        Parameters
        ----------
        params: str, optional
            | A string specifying which parameters to freeze.
            | Can contain any combination of:

            - `'s'` for initial state probabilities (HMM parameters),
            - `'t'` for transition probabilities (HMM parameters),
            - `'m'` for mean vectors (GMM emission distribution parameters),
            - `'c'` for covariance matrices (GMM emission distribution parameters),
            - '`w`' for mixing weights (GMM emission distribution parameters).

            Defaults to all parameters, i.e. `'stmcw'`.

        See Also
        --------
        unfreeze : Unfreezes parameters of the HMM or its GMM emission distributions.
        """
        self._frozen |= set(self._modify_params(params))

    def unfreeze(self, params=None):
        """Unfreezes the specified parameters of the HMM or its GMM emission distributions
        which were frozen with :func:`freeze`, allowing them to be updated during the Baum–Welch algorithm.

        Parameters
        ----------
        params: str, optional
            | A string specifying which parameters to unfreeze.
            | Can contain any combination of:

            - `'s'` for initial state probabilities (HMM parameters),
            - `'t'` for transition probabilities (HMM parameters),
            - `'m'` for mean vectors (GMM emission distribution parameters),
            - `'c'` for covariance matrices (GMM emission distribution parameters),
            - '`w`' for mixing weights (GMM emission distribution parameters).

            Defaults to all parameters, i.e. `'stmcw'`.

        See Also
        --------
        freeze : Freezes parameters of the HMM or its GMM emission distributions.
        """
        self._frozen -= set(self._modify_params(params))

    def _modify_params(self, params):
        if isinstance(params, str):
            if bool(re.compile(r'[^stmcw]').search(params)):
                raise ValueError("Expected a string consisting of any combination of 's', 't', 'm', 'c', 'w'")
        elif params is None:
            params = 'stmcw'
        else:
            raise TypeError("Expected a string consisting of any combination of 's', 't', 'm', 'c', 'w'")
        return params

    @property
    def label(self):
        return self._label

    @property
    def n_states(self):
        return self._n_states

    @property
    def n_components(self):
        return self._n_components

    @property
    def covariance_type(self):
        return self._covariance_type

    @property
    def n_seqs_(self):
        return self._val.is_fitted(self,
            lambda self: self._n_seqs_,
            'The model has not been fitted and has not seen any observation sequences'
        )

    @property
    def frozen(self):
        return self._frozen

    @property
    def model(self):
        return self._val.is_fitted(self,
            lambda self: self._model,
            'The model must be fitted first'
        )

    @property
    def initial_(self):
        return self._val.is_fitted(self,
            lambda self: self._initial_,
            'No initial state distribution has been defined'
        )

    @initial_.setter
    def initial_(self, probabilities):
        self._topology.validate_initial(probabilities)
        self._initial_ = probabilities

    @property
    def transitions_(self):
        return self._val.is_fitted(self,
            lambda self: self._transitions_,
            'No transition matrix has been defined'
        )

    @transitions_.setter
    def transitions_(self, probabilities):
        self._topology.validate_transitions(probabilities)
        self._transitions_ = probabilities

    @property
    def weights_(self):
        return self.model.weights_

    @property
    def means_(self):
        return self.model.means_

    @property
    def covars_(self):
        return self.model.covars_

    @property
    def monitor_(self):
        return self.model.monitor_

    def __repr__(self):
        name = '.'.join([self.__class__.__module__.split('.')[0], self.__class__.__name__])
        attrs = [
            ('label', repr(self._label)),
            ('n_states', repr(self._n_states)),
            ('n_components', repr(self._n_components)),
            ('covariance_type', repr(self._covariance_type)),
            ('frozen', repr(self._frozen))
        ]
        return '{}({})'.format(name, ', '.join('{}={}'.format(name, val) for name, val in attrs))


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from setuptools import setup, find_packages

import platform
from pkg_resources import packaging

VERSION = '0.12.1'

with open('README.md', 'r', encoding='utf8') as fh:
    long_description = fh.read()

pkg_versions = {
    'Cython': '>=0.28.5',
    'numpy': '>=1.17,<2',
    'scipy': '>=1.3,<2',
    'hmmlearn': '==0.2.4',
    'dtaidistance[numpy]': '>=2.2,<2.3',
    'scikit-learn': '>=0.22,<1',
    'tqdm': '>=4.36,<5',
    'joblib': '>=0.14,<1',
    # [torch]
    'torch': '>=1.8+cpu',
    # [docs]
    'sphinx': '==3.5.4',
    'numpydoc': '',
    'sphinx_rtd_theme': '',
    'm2r2': '',
    # [test]
    'pytest': '==5.3.2',
    # [notebooks]
    'jupyter': '==1.0.0',
    'requests': '==2.25.1',
    'matplotlib': '==3.3.3',
    'pandas': '==1.1.5',
    'seaborn': '==0.11.1',
    'torchaudio': '==0.8.0',
    'torchvision': '==0.9.0',
    'torchfsdd': '==0.1.2',
    'librosa': '==0.8.0'
}

extra_pkgs = {
    'torch': ['dev', 'torch', 'test', 'docs', 'notebooks'],
    'pytest': ['dev', 'test'],
    **{pkg:['dev', 'docs'] for pkg in ('sphinx', 'numpydoc', 'sphinx_rtd_theme', 'm2r2')},
    **{pkg:['dev', 'notebooks'] for pkg in (
        'jupyter', 'requests', 'matplotlib', 'pandas',
        'seaborn', 'tqdm', 'torchaudio', 'torchvision',
        'torchfsdd', 'librosa'
    )},
}

def load_requires(*pkgs):
    return [pkg + pkg_versions[pkg] for pkg in pkgs]

def reverse_extra(extra):
    return [pkg + pkg_versions[pkg] for pkg, extras in extra_pkgs.items() if extra in extras]

python_requires = '>=3.6,<3.10'

setup_requires = load_requires('Cython', 'numpy', 'scipy')

install_requires = load_requires('numpy', 'hmmlearn', 'dtaidistance[numpy]', 'scipy', 'scikit-learn', 'tqdm', 'joblib')
if packaging.version.parse(platform.python_version()) < packaging.version.parse('3.8'):
    install_requires.append('importlib_metadata') # Backports for importlib.metadata in Python <3.8

extras_require = {extra:reverse_extra(extra) for extra in ('torch', 'dev', 'test', 'notebooks', 'docs')}

setup(
    name = 'sequentia',
    version = VERSION,
    author = 'Edwin Onuonga',
    author_email = 'ed@eonu.net',
    description = 'A machine learning interface for isolated sequence classification algorithms in Python.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/eonu/sequentia',
    project_urls = {
        'Documentation': 'https://sequentia.readthedocs.io/en/latest',
        'Bug Tracker': 'https://github.com/eonu/sequentia/issues',
        'Source Code': 'https://github.com/eonu/sequentia'
    },
    license = 'MIT',
    package_dir = {'': 'lib'},
    packages = find_packages(where='lib'),
    classifiers = [
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Natural Language :: English'
    ],
    python_requires = python_requires,
    setup_requires = setup_requires,
    install_requires = install_requires,
    extras_require = extras_require
)
